#ifndef __LINES_OF_ACTION_HPP__
#define __LINES_OF_ACTION_HPP__

#include "game.hpp"
#include <random>
#include <array>
#include <vector>
#include <iostream>
#include <memory>

namespace game
{

  struct lines_of_action_state
  {
    uint64_t black_bitboard = 0x7E0000000000007E;
    uint64_t white_bitboard = 0x0081818181818100;
    uint32_t row_nb_pieces = 0x62222226;
    uint32_t column_nb_pieces = 0x62222226;
    uint64_t left_to_right_diagonal_nb_pieces = 0x0022222202222220;
    uint64_t right_to_left_diagonal_nb_pieces = 0x0022222202222220;
    uint64_t hash_value = 0;
    uint16_t nb_moves = 48;
    uint8_t black_nb_pieces = 12;
    uint8_t white_nb_pieces = 12;
    uint8_t depth = 0;
    bool first_player_win = false;
    bool second_player_win = false;
  };

  class lines_of_action : public game<lines_of_action_state>
  {
  public:
    lines_of_action() = default;
    lines_of_action(const lines_of_action& loa) = default;
    lines_of_action& operator=(const lines_of_action& loa) = default;
    bool end_of_game() const;
    int value(std::uint8_t player) const;
    bool won(std::uint8_t player) const;
    bool lost(std::uint8_t player) const;
    bool draw(std::uint8_t player) const;
    uint8_t current_player() const;
    std::uint16_t number_of_moves() const;
    void play(std::uint16_t m);
    void undo(std::uint16_t m) {}
    std::string player_to_string(std::uint8_t player) const;
    std::string move_to_string(std::uint16_t m) const;
    std::string to_string() const;
    void playout(std::mt19937& engine, int max_depth = -1);
    lines_of_action_state get_state();
    void set_state(const lines_of_action_state& state);
    std::shared_ptr<game<lines_of_action_state>> do_copy() const;
    std::uint64_t hash(std::uint16_t m) const;
    std::uint64_t hash() const;
  private:    
    const uint8_t BLACK = 0;
    const uint8_t WHITE = 1;
    const uint8_t EMPTY = 2;
    struct move
    {
      uint8_t from_i, from_j;
      uint8_t to_i, to_j;
      move(uint8_t fi, uint8_t fj, uint8_t ti, uint8_t tj) : from_i(fi), from_j(fj), to_i(ti), to_j(tj) {}
      std::string to_string() const
      {
	return char('a' + from_j) + std::to_string(8 - from_i) + "-" +  
	  char('a' + to_j) + std::to_string(8 - to_i);
      }
    };
    lines_of_action_state state;
    inline uint8_t get(uint8_t i, uint8_t j) const;
    inline bool is_valid_row_right(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const;
    inline bool is_valid_row_left(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const;
    inline bool is_valid_column_up(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const;
    inline bool is_valid_column_down(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const;
    inline bool is_valid_left_to_right_diagonal_up(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const;
    inline bool is_valid_left_to_right_diagonal_down(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const;
    inline bool is_valid_right_to_left_diagonal_up(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const;
    inline bool is_valid_right_to_left_diagonal_down(uint64_t bitboard1, uint64_t bitboard2, uint8_t from_i, uint8_t from_j, uint8_t n) const;
    move move_index_to_move(uint16_t m) const;
    void update_state(const move& m);
    static std::vector<std::vector<uint64_t>> black_hash_values;
    static std::vector<std::vector<uint64_t>> white_hash_values;
  };
  std::ostream& operator<<(std::ostream& os, const lines_of_action& loa);
}

#endif
